#!/usr/bin/env python
####
#
#  Copyright (c) 2008-2012 Aerospike, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Description: Utility to monitor the status of the nodes and send commands to the cluster
# 
# 
####






import sys, os, time, atexit
from signal import SIGTERM
import fcntl
import subprocess


class Pidfile(object):
	def __init__(self, pidfile, procname):
		try:
			self.fd = os.open(pidfile, os.O_CREAT | os.O_RDWR)
		except IOError, e:
			sys.exit("Failed to open pidfile: %s" % str(e))
		self.pidfile = pidfile
		self.procname = procname
		assert not fcntl.flock(self.fd, fcntl.LOCK_EX)

	def unlock(self):
		assert not fcntl.flock(self.fd, fcntl.LOCK_UN)

	def write(self, pid):
		os.ftruncate(self.fd, 0)
		os.write(self.fd, "%d" % int(pid))
		os.fsync(self.fd)

	def kill(self):
		pid = int(os.read(self.fd, 4096))
		os.lseek(self.fd, 0, os.SEEK_SET)
		
		try:
			os.kill(pid, SIGTERM)
			time.sleep(0.1)
		except OSError, err:
			err = str(err)
			if err.find("No such process") > 0:
					os.remove(self.pidfile)
			else:
				return str(err)

		if self.is_running():
			return "Failed to kill %d" % pid

	def is_running(self):
		contents = os.read(self.fd, 4096)
		os.lseek(self.fd, 0, os.SEEK_SET)

		if not contents:
			return False

		p = subprocess.Popen(["ps", "-o", "comm", "-p", str(int(contents))],
				stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		stdout, stderr = p.communicate()
		if stdout == "COMM\n":
			return False

		if self.procname in stdout[stdout.find("\n") + 1:]:
			return True

		return False


class Daemon:
	"""
	A generic daemon class.
	Usage: subclass the Daemon class and override the run() method
	"""
	def __init__(self, pidfile, logfile , stdin='/dev/null'):
		self.stdin = stdin
		self.stdout = logfile
		self.stderr = logfile
		self.pidfile = Pidfile(pidfile, "python")

	def daemonize(self):
		"""
		do the UNIX double-fork magic, see Stevens' "Advanced 
		Programming in the UNIX Environment" for details (ISBN 0201563177)
		http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
		"""
		try: 
			pid = os.fork() 
			if pid > 0:
				# exit first parent
				sys.exit(0) 
		except OSError, e: 
			sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
			sys.exit(1)
	
		# decouple from parent environment
		os.chdir("/") 
		os.setsid() 
		os.umask(0) 
	
		# do second fork
		try: 
			pid = os.fork() 
			if pid > 0:
				# exit from second parent
				sys.exit(0) 
		except OSError, e: 
			sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
			sys.exit(1) 
	
		# redirect standard file descriptors
		sys.stdout.flush()
		sys.stderr.flush()
		si = file(self.stdin, 'r')
		so = file(self.stdout, 'a+')
		se = file(self.stderr, 'a+', 0)
		os.dup2(si.fileno(), sys.stdin.fileno())
		os.dup2(so.fileno(), sys.stdout.fileno())
		os.dup2(se.fileno(), sys.stderr.fileno())
	
		# write pidfile
		atexit.register(self.delpid)
		pid = str(os.getpid())
		self.pidfile.write(pid)
		
	def delpid(self):
		os.remove(self.pidfile)

	def start(self):
		"""
		Start the daemon
		"""
		# Check for a pidfile to see if the daemon already runs
		if self.pidfile.is_running():
			self.pidfile.unlock()
			sys.exit("Daemon already running.")

		# Start the daemon
		self.daemonize()
		self.pidfile.unlock()
		self.run()

	def stop(self):
		"""
		Stop the daemon
		"""
		# Get the pid from the pidfile
		if not self.pidfile.is_running():
			self.pidfile.unlock()
			print >> sys.stderr, "Daemon not running."
			return

		# Try killing the daemon process
		error = self.pidfile.kill()
		if error:
			self.pidfile.unlock()
			sys.exit(error)

	def restart(self):
		"""
		Restart the daemon
		"""
		self.stop()
		self.start()
	
	def run(self):
		"""
		You should override this method when you subclass Daemon. It will be called after the process has been
		daemonized by start() or restart().
		"""
